class StoryError(Exception):
    pass


class StoryDefinitionError(StoryError):
    pass


class FailureError(StoryError):
    def __init__(self, reason):
        self.__reason = reason
        super(FailureError, self).__init__()

    def __repr__(self):
        reason = repr(self.__reason) if self.__reason else ""
        return "FailureError(" + reason + ")"


class FailureProtocolError(StoryError):
    pass


class ContextContractError(StoryError):
    pass


class MutationError(StoryError):
    pass



# type: ignore
from raven.breadcrumbs import libraryhook
from raven.breadcrumbs import record

import _stories.context


# FIXME: Test me.
#
# FIXME: Type me.


origin_context_init = _stories.context.Context.__init__


@libraryhook("stories")
def track_context():
    def wrapper(ctx):
        origin_context_init(ctx)
        record(
            processor=lambda data: data.update(
                {"category": "story", "message": repr(ctx)}
            )
        )

    _stories.context.Context.__init__ = wrapper



from enum import Enum
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

from _stories.contract import NullContract
from _stories.contract import SpecContract
from _stories.failures import NotNullExecProtocol
from _stories.failures import NullExecProtocol
from _stories.marker import BeginningOfStory
from _stories.marker import EndOfStory
from _stories.summary import FailureSummary
from _stories.summary import SuccessSummary

class ClassMountedStory:
    def __init__(
        self,
        cls: Any,
        name: str,
        collected: List[str],
        contract: Callable[[Any], Any],
        failures: Callable[[Any], Optional[Union[List[str], Type[Enum]]]],
    ) -> None: ...
    def __repr__(self) -> str: ...

class MountedStory:
    def __init__(
        self,
        obj: Any,
        cls_name: str,
        name: str,
        arguments: List[str],
        methods: List[
            Tuple[
                Union[BeginningOfStory, Callable, EndOfStory],
                Union[NullContract, SpecContract],
                Union[NullExecProtocol, NotNullExecProtocol],
            ],
        ],
        contract: NullContract,
        failures: Optional[Union[List[str], Type[Enum]]],
    ) -> None: ...
    def __call__(self, **kwargs: Dict[str, Any]) -> Optional[Union[List[str], int]]: ...
    def run(
        self, **kwargs: Dict[str, Any]
    ) -> Union[SuccessSummary, FailureSummary]: ...
    def __repr__(self) -> str: ...



"""
stories.contrib.sentry
----------------------

This module contains integration with Sentry.

:copyright: (c) 2018-2020 dry-python team.
:license: BSD, see LICENSE for more details.
"""
import _stories.contrib.sentry  # noqa


__all__ = []



# type: ignore
"""
stories.contrib.debug_toolbars.django
-------------------------------------

This module contains integration with Django Debug Toolbar.

:copyright: (c) 2018-2020 dry-python team.
:license: BSD, see LICENSE for more details.
"""
from _stories.contrib.debug_toolbars.django.panels import StoriesPanel


__all__ = ["StoriesPanel"]



from enum import Enum
from typing import Any
from typing import Optional
from typing import Union

class History:
    def __init__(self) -> None: ...
    def before_call(self, method_name: str) -> None: ...
    def on_result(self, value: Any) -> None: ...
    def on_failure(self, reason: Optional[Union[str, Enum]]) -> None: ...
    def on_skip(self) -> None: ...
    def on_error(self, error_name: str) -> None: ...
    def on_substory_start(self) -> None: ...
    def on_substory_end(self) -> None: ...



"""
stories
-------

This module implements Business Transaction DSL.

:copyright: (c) 2018-2020 dry-python team.
:license: BSD, see LICENSE for more details.
"""
from _stories.argument import arguments
from _stories.returned import Failure
from _stories.returned import Result
from _stories.returned import Skip
from _stories.returned import Success
from _stories.story import Story as story


__all__ = ["story", "arguments", "Result", "Success", "Failure", "Skip"]



class Result(object):
    def __init__(self, value=None):
        self.value = value

    def __repr__(self):
        return "Result(" + repr(self.value) + ")"


class Success(object):
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __repr__(self):
        return (
            "Success("
            + ", ".join([k + "=" + repr(v) for k, v in self.kwargs.items()])
            + ")"
        )


class Failure(object):
    def __init__(self, reason=None):
        self.reason = reason

    def __repr__(self):
        reason = repr(self.reason) if self.reason else ""
        return "Failure(" + reason + ")"


class Skip(object):
    def __repr__(self):
        return "Skip()"



from _stories.argument import get_arguments
from _stories.collect import collect_story
from _stories.failures import check_data_type
from _stories.mounted import ClassMountedStory
from _stories.mounted import MountedStory
from _stories.wrap import wrap_story


class Story(object):
    def __init__(self, f):
        self.name = f.__name__
        self.arguments = get_arguments(f)
        self.collected = collect_story(f)
        self.contract(None)
        self.failures(None)

    def __get__(self, obj, cls):
        __tracebackhide__ = True
        if obj is None:
            return ClassMountedStory(
                cls, self.name, self.collected, self.contract, self.failures
            )
        else:
            methods, contract, failures = wrap_story(
                self.arguments,
                self.collected,
                cls.__name__,
                self.name,
                obj,
                self.__contract,
                self.__failures,
            )
            return MountedStory(
                obj,
                cls.__name__,
                self.name,
                self.arguments,
                methods,
                contract,
                failures,
            )

    def contract(self, contract):
        # FIXME: Raise error on unsupported types.
        self.__contract = contract
        return contract

    def failures(self, failures):
        check_data_type(failures)
        self.__failures = failures
        return failures



# NOTE: Every type ignore in this module is necessary due to
# false-positive bugs in mypy.  Feel free to remove them as they'll be
# fixed.  See GitHub issues for more info:
# `https://github.com/python/mypy/issues/1105`,
# `https://github.com/python/mypy/issues/1106`,
# `https://github.com/python/mypy/issues/1107`.


try:
    from enum import Enum, EnumMeta
except ImportError:
    # We are on Python 2.7 and enum34 package is not installed.
    class Enum(object):  # type: ignore
        pass

    class EnumMeta(object):  # type: ignore
        pass


try:
    from pydantic.error_wrappers import ErrorWrapper as PydanticError
    from pydantic.main import ModelMetaclass as PydanticSpec
except ImportError:
    # Pydantic package is not installed.
    class PydanticSpec(object):  # type: ignore
        pass

    class PydanticError(object):  # type: ignore
        pass


try:
    from marshmallow.schema import SchemaMeta as MarshmallowSpec
except ImportError:
    # Marshmallow package is not installed.
    class MarshmallowSpec(object):  # type: ignore
        pass


try:
    from cerberus import Validator as CerberusSpec
except ImportError:
    # Cerberus package is not installed.
    class CerberusSpec(object):  # type: ignore
        pass


try:
    from textwrap import indent
except ImportError:
    # We are on Python 2.7
    def indent(text, prefix):  # type: ignore
        return "".join(map(lambda l: prefix + l, text.splitlines(True)))


try:
    from prettyprinter import pformat
except ImportError:
    # Prettyprinter package is not installed.
    from pprint import pformat  # noqa



from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import overload
from typing import Set
from typing import Tuple
from typing import Type
from typing import Union

from cerberus.validator import Validator
from marshmallow.schema import Schema
from pydantic.error_wrappers import ErrorWrapper
from pydantic.fields import ModelField
from pydantic.main import BaseModel
from typing_extensions import Literal

from _stories.context import Context
from _stories.failures import NotNullExecProtocol
from _stories.failures import NullExecProtocol
from _stories.marker import BeginningOfStory
from _stories.marker import EndOfStory
from _stories.returned import Failure
from _stories.returned import Result
from _stories.returned import Skip
from _stories.returned import Success

class PydanticValidator:
    def __init__(self, spec: BaseModel, field: ModelField) -> None: ...
    def __call__(self, value: Any) -> Any: ...
    def __repr__(self) -> str: ...

class MarshmallowValidator:
    def __init__(self, spec: Schema, field: str) -> None: ...
    def __call__(self, value: Any) -> Any: ...
    def __repr__(self) -> str: ...

class CerberusValidator:
    def __init__(self, spec: Validator, field: str) -> None: ...
    def __call__(self, value: Any) -> Any: ...
    def __repr__(self) -> str: ...

class RawValidator:
    def __init__(self, validator: Callable) -> None: ...
    def __call__(self, value: Any) -> Any: ...
    def __repr__(self) -> str: ...

def disassemble_pydantic(spec: BaseModel) -> Dict[str, PydanticValidator]: ...
def disassemble_marshmallow(spec: Schema) -> Dict[str, MarshmallowValidator]: ...
def disassemble_cerberus(spec: Validator) -> Dict[str, CerberusValidator]: ...
def disassemble_raw(spec: Dict[str, Callable]) -> Dict[str, RawValidator]: ...
@overload
def make_contract(
    cls_name: str, name: str, arguments: List[str], spec: Literal[None]
) -> NullContract: ...
@overload
def make_contract(
    cls_name: str, name: str, arguments: List[str], spec: Type[BaseModel]
) -> SpecContract: ...
@overload
def make_contract(
    cls_name: str, name: str, arguments: List[str], spec: Type[Schema]
) -> SpecContract: ...
@overload
def make_contract(
    cls_name: str, name: str, arguments: List[str], spec: Dict[str, Callable]
) -> SpecContract: ...
def check_arguments_definitions(
    cls_name: str,
    name: str,
    arguments: List[str],
    spec: Dict[
        str,
        Union[MarshmallowValidator, PydanticValidator, RawValidator, CerberusValidator],
    ],
) -> None: ...

class NullContract:
    def __init__(self, cls_name: str, name: str, arguments: List[str]) -> None: ...
    def set_null(self) -> None: ...
    def make_argset(self) -> None: ...
    def check_story_call(self, kwargs: Dict[str, Any]) -> Dict[str, Any]: ...
    def check_substory_call(self, ctx: Context) -> None: ...
    def check_success_statement(
        self, method: Callable, ctx: Context, ns: Dict[str, Any]
    ) -> Dict[str, Any]: ...
    def __repr__(self) -> str: ...
    def format_contract_fields(
        self, *fieldset: Tuple[Dict[str, Set[Tuple[None, str, str]]], ...]
    ) -> str: ...

class SpecContract(NullContract):  # FIXME: Generic.
    def __init__(
        self,
        cls_name: str,
        name: str,
        arguments: List[str],
        spec: Dict[
            str,
            Union[
                MarshmallowValidator, PydanticValidator, RawValidator, CerberusValidator
            ],
        ],
        origin: Any,
    ) -> None: ...
    def set_null(self) -> None: ...
    def make_argset(self) -> None: ...
    def make_declared(self) -> None: ...
    def check_story_call(
        self, kwargs: Dict[str, Union[Dict[str, str], str, int, List[int], List[str]]]
    ) -> Dict[str, Union[int, List[int], Dict[str, str], Dict[str, int]]]: ...
    def check_success_statement(
        self,
        method: Callable[[Context], Union[Result, Success, Failure, Skip]],
        ctx: Context,
        ns: Dict[str, Any],
    ) -> Dict[str, Any]: ...
    def identify(self, ns: Dict[str, Any]) -> Set[str]: ...
    def validate(self, ns: Dict[str, Any]) -> Any: ...
    def validate_spec(
        self,
        result: Dict[str, Union[Dict[str, str], int]],
        errors: Dict[str, Union[List[str], ErrorWrapper, str]],
        seen: Union[List[Tuple[str, str]], List[Tuple[str, Dict[str, str]]]],
        key: str,
        value: Union[List[str], Dict[str, str], str],
    ) -> None: ...
    def validate_argset(
        self,
        result: Dict[str, Union[Dict[str, str], str, int, List[int]]],
        errors: Dict[str, Union[List[str], ErrorWrapper, str]],
        seen: Union[
            List[Tuple[str, str]],
            List[Union[Tuple[str, str], Tuple[str, List[str]]]],
            List[Tuple[str, Dict[str, str]]],
        ],
        conflict: Dict[Tuple[str, str], Dict[str, Union[int, str]]],
        key: str,
        value: Union[List[int], List[str], Dict[str, str], str],
    ) -> None: ...
    def assign_result(
        self,
        result: Dict[str, Union[Dict[str, str], str, int, List[int]]],
        seen: Union[
            List[Tuple[str, str]],
            List[Union[Tuple[str, str], Tuple[str, List[str]]]],
            List[Tuple[str, Dict[str, str]]],
        ],
        key: str,
        value: Union[List[int], str, Dict[str, str], List[str]],
        new_value: Any,
    ) -> None: ...
    def __repr__(self) -> str: ...
    def format_contract_fields(
        self, *fieldset: Tuple[Dict[str, Set[Tuple[None, str, str]]], ...]
    ) -> str: ...

def format_violations(ns: Dict[str, Any], errors: Dict[str, Any]) -> str: ...
def combine_contract(
    parent: Union[SpecContract, NullContract], child: Union[SpecContract, NullContract]
) -> None: ...
def combine_argsets(
    parent: Union[SpecContract, NullContract], child: Union[SpecContract, NullContract]
) -> None: ...
def combine_declared(parent: SpecContract, child: SpecContract) -> None: ...
def format_contract(
    contract: Union[SpecContract, NullContract]
) -> Optional[Union[Type[BaseModel], Type[Schema], Type[dict], Type[Validator]]]: ...
def maybe_extend_downstream_argsets(
    methods: List[
        Tuple[
            Union[BeginningOfStory, Callable, EndOfStory],
            Union[NullContract, SpecContract],
            Union[NullExecProtocol, NotNullExecProtocol],
        ],
    ],
    root: Union[SpecContract, NullContract],
) -> None: ...



from _stories.contract import combine_contract
from _stories.contract import make_contract
from _stories.contract import maybe_extend_downstream_argsets
from _stories.failures import combine_failures
from _stories.failures import make_exec_protocol
from _stories.failures import maybe_disable_null_protocol
from _stories.marker import BeginningOfStory
from _stories.marker import EndOfStory
from _stories.mounted import MountedStory


def wrap_story(arguments, collected, cls_name, story_name, obj, spec, failures):
    __tracebackhide__ = True

    contract = make_contract(cls_name, story_name, arguments, spec)
    protocol = make_exec_protocol(failures)

    methods = [(BeginningOfStory(cls_name, story_name), contract, protocol)]

    for name in collected:

        attr = getattr(obj, name)

        if type(attr) is not MountedStory:
            methods.append((attr, contract, protocol))
            continue

        combine_contract(contract, attr.contract)

        failures = combine_failures(
            failures, cls_name, story_name, attr.failures, attr.cls_name, attr.name
        )

        # FIXME: Is there a way to avoid this modification?
        attr.methods[0][0].set_parent(name, attr.obj is obj)

        methods.extend(attr.methods)

    methods.append((EndOfStory(is_empty=not collected), contract, protocol))

    maybe_extend_downstream_argsets(methods, contract)

    methods = maybe_disable_null_protocol(methods, failures)

    return methods, contract, failures



"""
stories.contrib.sentry.django
-----------------------------

This module contains integration with Sentry for Django framework.

:copyright: (c) 2018-2020 dry-python team.
:license: BSD, see LICENSE for more details.
"""
from raven.contrib.django.client import DjangoClient


__all__ = ["DjangoClient"]



# type: ignore
from flask import render_template
from flask_debugtoolbar.panels import DebugPanel

import _stories.context


# FIXME: Test me.
#
# FIXME: Type me.


original_context_init = _stories.context.Context.__init__


def track_context(storage):
    def wrapper(ctx):
        original_context_init(ctx)
        storage.append(ctx)

    return wrapper


def pluralize(number, singular, plural=None):
    if plural is None:
        plural = singular + "s"
    if number == 1:
        return "%d %s" % (number, singular)
    else:
        return "%d %s" % (number, plural)


class StoriesPanel(DebugPanel):
    name = "Stories"
    has_content = True

    def __init__(self, *args, **kwargs):
        super(StoriesPanel, self).__init__(*args, **kwargs)
        self.storage = []
        self.enable_instrumentation()

    def nav_title(self):
        return "Stories"

    def nav_subtitle(self):
        count = len(self.storage)
        return pluralize(count, "call")

    def title(self):
        count = len(self.storage)
        return "Context and execution path of %s" % pluralize(count, "story", "stories")

    def url(self):
        return "#"

    def content(self):
        return render_template(
            "stories/debug_toolbar/stories_panel.html", stories=self.storage
        )

    def enable_instrumentation(self):
        _stories.context.Context.__init__ = track_context(self.storage)

    def disable_instrumentation(self):
        _stories.context.Context.__init__ = original_context_init



from enum import Enum
from typing import Any
from typing import Callable
from typing import List
from typing import NoReturn
from typing import Optional
from typing import overload
from typing import Tuple
from typing import Type
from typing import Union

from _stories.contract import NullContract
from _stories.contract import SpecContract
from _stories.marker import BeginningOfStory
from _stories.marker import EndOfStory

def check_data_type(failures: Any) -> None: ...
def failures_representation(failures: Union[List[str], Type[Enum]]) -> str: ...
def collection_contains(reason: str, failures: List[str]) -> bool: ...
def collection_compare(a: str, b: str) -> bool: ...
def enumeration_contains(reason: Enum, failures: Type[Enum]) -> bool: ...
def enumeration_compare(a: Enum, b: Enum) -> bool: ...
@overload
def make_exec_protocol(failures: None) -> NullExecProtocol: ...
@overload
def make_exec_protocol(failures: List[str]) -> NotNullExecProtocol: ...
@overload
def make_exec_protocol(failures: Type[Enum]) -> NotNullExecProtocol: ...

class NullExecProtocol:
    def check_return_statement(
        self, method: Callable, reason: Optional[Union[str, Enum]]
    ) -> None: ...

class DisabledNullExecProtocol:
    def check_return_statement(
        self, method: Callable, reason: Optional[Union[str, Enum]]
    ) -> None: ...

class NotNullExecProtocol:  # FIXME: Generic.
    def __init__(
        self,
        failures: Union[List[str], Type[Enum], Type[Enum]],
        contains_func: Callable,
    ) -> None: ...
    def check_return_statement(
        self, method: Callable, reason: Optional[Union[str, Enum]]
    ) -> None: ...

@overload
def make_run_protocol(
    failures: None, cls_name: str, method_name: str
) -> NullRunProtocol: ...
@overload
def make_run_protocol(
    failures: List[str], cls_name: str, method_name: str
) -> NotNullRunProtocol: ...
@overload
def make_run_protocol(
    failures: Type[Enum], cls_name: str, method_name: str
) -> NotNullRunProtocol: ...

class NullRunProtocol:
    def __init__(self, cls_name: str, method_name: str) -> None: ...
    def check_failed_because_argument(self, reason: str) -> NoReturn: ...

class NotNullRunProtocol:  # FIXME: Generic.
    def __init__(
        self,
        cls_name: str,
        method_name: str,
        failures: Union[List[str], Type[Enum]],
        contains_func: Callable,
        compare_func: Callable,
    ) -> None: ...
    def check_failed_because_argument(self, reason: Union[str, Enum]) -> None: ...
    def compare_failed_because_argument(
        self, argument: Union[str, Enum], failure_reason: Union[str, Enum]
    ) -> bool: ...

def combine_failures(
    first_failures: Optional[Union[List[str], Type[Enum]]],
    first_cls_name: str,
    first_method_name: str,
    second_failures: Optional[Union[List[str], Type[Enum]]],
    second_cls_name: str,
    second_method_name: str,
) -> Optional[Union[List[str], Type[Enum]]]: ...
def maybe_disable_null_protocol(
    methods: List[
        Tuple[
            Union[BeginningOfStory, Callable, EndOfStory],
            Union[NullContract, SpecContract],
            Union[NullExecProtocol, NotNullExecProtocol],
        ],
    ],
    reasons: Optional[Union[List[str], Type[Enum]]],
) -> List[
    Tuple[
        Union[BeginningOfStory, Callable, EndOfStory],
        Union[NullContract, SpecContract],
        Union[NullExecProtocol, DisabledNullExecProtocol, NotNullExecProtocol],
    ],
]: ...



from inspect import isclass
from operator import itemgetter

from _stories.compat import CerberusSpec
from _stories.compat import MarshmallowSpec
from _stories.compat import PydanticError
from _stories.compat import PydanticSpec
from _stories.exceptions import ContextContractError


# FIXME: Handle protocol extension.  There should be way to say in the
# substory contract "this variable should be an integer" and in
# addition in the story "this integer should be greater then 7".  This
# way we also can require a certain substory to declare context
# variable for parent story.
#
# FIXME: Add fix suggestion to the bottom of the error message.
#
# FIXME: Support custom validation of complex relationships.  For
# example, `@pydantic.validator` of `password2` depends on
# `password1`.
#
# FIXME: Check that contracts collisions are checked in the situation:
#
# story
#   substory
#     substory with variable declaration
#   substory with this variable as an argument
#
# FIXME: Depending of the level of nesting the same substory can
# occurred multiple times at one argument in the contract
# representation.
#
# defaults:
#   typing.Union[typing.Dict[str, str], NoneType]  # Argument of ProcessImages.process
#   typing.Union[typing.Dict[str, str], NoneType]  # Argument of ProcessImages.process
#   typing.Union[typing.Dict[str, str], NoneType]  # Argument of ProcessImages.process
# raw_urls:
#   str  # Argument of FetchURLPreviews.fetch
#   str  # Argument of FetchURLPreviews.fetch
# files:
#   typing.Dict[str, typing.Union[str, int, typing.Any]]  # Argument of ProcessVideos.process  # noqa: E501
#   typing.Dict[str, typing.Union[str, int, typing.Any]]  # Argument of ProcessImages.process  # noqa: E501
#   typing.Dict[str, typing.Union[str, int, typing.Any]]  # Argument of ProcessImages.process  # noqa: E501
#   typing.Dict[str, typing.Union[str, int, typing.Any]]  # Argument of ProcessImages.process  # noqa: E501
#
# FIXME: Fix pydantic error messages.
#
# In [1]: class Context(pydantic.BaseModel):
#    ...:     files: typing.Dict[str, typing.Dict[str, typing.Union[str, int, typing.BinaryIO]]]  # noqa: E501
#    ...:
#
# In [2]: Context(files={"a": {'name': B(name='test'), 'size': 1}})
# ---------------------------------------------------------------------------
# ValidationError: 3 validation errors
# files -> a -> name
#   str type expected (type=type_error.str)
# files -> a -> name
#   value is not a valid integer (type=type_error.integer)
# files -> a -> name
#   instance of BinaryIO expected (type=type_error.arbitrary_type; expected_arbitrary_type=BinaryIO)  # noqa: E501
#
# FIXME: Test all field representation for all supported libraries.
# I.e. Dict, List, Tuple, Integer, String, etc.
#
# FIXME: Alias validation creates new object for each alias.  So they
# became not aliases.
#
# FIXME: Parent story can't define an argument, if child story already
# defined variable with the same name.
#
# FIXME: When Success argument is broken we should show validator of
# what argument of what substory is broken.  Because it's really hard
# to track down it in a deep story composition.
#
# FIXME: Nested collision detector doesn't support identity check for
# neighbor substories.  Only parent-child for now.
#
# class Subs
#   @story
#   def foo(I):
#
#   @story
#   def bar(I):
#
#   @contract_in(Subs)
#   class Context:
#     bar: int
#
# class Parent:
#   @story
#   def quiz(I):
#     I.foo
#     I.bar   <--- `bar` will be treated as repeated variable.
#
# NOTE: `noqa` comments are not part of the program output.


# Validators.


class PydanticValidator(object):
    def __init__(self, spec, field):
        self.spec = spec
        self.field = field

    def __call__(self, value):
        return self.field.validate(value, {}, loc=self.field.alias, cls=self.spec)

    def __repr__(self):
        return self.field._type_display()


class MarshmallowValidator(object):
    def __init__(self, spec, field):
        self.spec = spec
        self.field = field

    def __call__(self, value):
        values, errors = self.spec().load({self.field: value})
        return values.get(self.field), errors.get(self.field)

    def __repr__(self):
        field = self.spec._declared_fields[self.field]
        return field.__class__.__name__


class CerberusValidator(object):
    def __init__(self, spec, field):
        self.spec = spec
        self.field = field

    def __call__(self, value):
        validated = CerberusSpec()
        validated.validate({self.field: value}, self.spec.schema.schema)
        return validated.document.get(self.field), validated.errors.get(self.field)

    def __repr__(self):
        schema = self.spec.schema.schema[self.field]
        field_type = schema["type"]
        if "schema" in schema and "type" in schema["schema"]:
            field_type += "[" + schema["schema"]["type"] + "]"
        return field_type


class RawValidator(object):
    def __init__(self, validator):
        self.validator = validator

    def __call__(self, value):
        return self.validator(value)

    def __repr__(self):
        return self.validator.__name__


# Disassemble.


def disassemble_pydantic(spec):
    result = {}
    for name, field in spec.__fields__.items():
        result[name] = PydanticValidator(spec, field)
    return result


def disassemble_marshmallow(spec):
    result = {}
    for name in spec._declared_fields:
        result[name] = MarshmallowValidator(spec, name)
    return result


def disassemble_cerberus(spec):
    result = {}
    for name in spec.schema:
        result[name] = CerberusValidator(spec, name)
    return result


def disassemble_raw(spec):
    result = {}
    for name, validator in spec.items():
        result[name] = RawValidator(validator)
    return result


# Execute.


def make_contract(cls_name, name, arguments, spec):
    __tracebackhide__ = True
    if spec is None:
        return NullContract(cls_name, name, arguments)
    elif isinstance(spec, PydanticSpec):
        disassembled = disassemble_pydantic(spec)
    elif isinstance(spec, MarshmallowSpec):
        disassembled = disassemble_marshmallow(spec)
    elif isinstance(spec, CerberusSpec):
        disassembled = disassemble_cerberus(spec)
    elif isinstance(spec, dict):
        disassembled = disassemble_raw(spec)
    check_arguments_definitions(cls_name, name, arguments, disassembled)
    return SpecContract(cls_name, name, arguments, disassembled, spec)


def check_arguments_definitions(cls_name, name, arguments, spec):
    __tracebackhide__ = True
    undefined = set(arguments) - set(spec)
    if undefined:
        message = undefined_argument_template.format(
            undefined=", ".join(sorted(undefined)),
            cls=cls_name,
            method=name,
            arguments=", ".join(arguments),
        )
        raise ContextContractError(message)


class NullContract(object):
    def __init__(self, cls_name, name, arguments):
        self.cls_name = cls_name
        self.name = name
        self.arguments = arguments
        self.make_argset()

    def make_argset(self):
        self.argset = {
            arg: {(None, self.cls_name, self.name)} for arg in self.arguments
        }

    def check_story_call(self, kwargs):
        __tracebackhide__ = True
        # FIXME: Check required arguments here.
        unknown_arguments = set(kwargs) - set(self.argset)
        if unknown_arguments:
            message = unknown_argument_template.format(
                unknown=", ".join(sorted(unknown_arguments)),
                cls=self.cls_name,
                method=self.name,
                contract=self,
            )
            raise ContextContractError(message)
        return kwargs

    def check_substory_call(self, ctx):
        __tracebackhide__ = True
        missed = set(self.arguments) - set(ctx._Context__ns)
        if missed:
            message = missed_variable_template.format(
                missed=", ".join(sorted(missed)),
                cls=self.cls_name,
                method=self.name,
                arguments=", ".join(self.arguments),
                ctx=ctx,
            )
            raise ContextContractError(message)

    def check_success_statement(self, method, ctx, ns):
        __tracebackhide__ = True
        tries_to_override = set(ctx._Context__ns) & set(ns)
        if tries_to_override:
            message = variable_override_template.format(
                variables=", ".join(map(repr, sorted(tries_to_override))),
                cls=method.__self__.__class__.__name__,
                method=method.__name__,
                ctx=ctx,
            )
            raise ContextContractError(message)
        return ns

    def __repr__(self):
        return self.format_contract_fields(self.argset)

    def format_contract_fields(self, *fieldset):
        if not self.argset:
            return "Contract()"
        lines = ["Contract:"]
        arguments = sorted(
            [field for fields in fieldset for field in fields if field in self.argset]
        )
        for argument in arguments:
            # FIXME: This does not work for story composition when
            # many stories has the same argument.
            ((validator, cls_name, name),) = self.argset[argument]
            lines.append("  %s  # Argument of %s.%s" % (argument, cls_name, name))
        return "\n".join(lines)


class SpecContract(NullContract):
    # FIXME: Deny empty disassembled spec.  If there is such need, we
    # should replace `NullContract` with empty `SpecContract`
    # ourselves.
    def __init__(self, cls_name, name, arguments, spec, origin):
        self.spec = spec
        self.origin = origin
        super(SpecContract, self).__init__(cls_name, name, arguments)
        self.make_declared()

    def make_argset(self):
        self.argset = {}
        for arg in self.arguments:
            self.argset[arg] = {(self.spec[arg], self.cls_name, self.name)}
            del self.spec[arg]

    def make_declared(self):
        self.declared = {
            variable: (self.cls_name, self.name, repr(validator))
            for variable, validator in self.spec.items()
        }

    def check_story_call(self, kwargs):
        __tracebackhide__ = True
        super(SpecContract, self).check_story_call(kwargs)
        result, errors = self.validate(kwargs)
        if errors:
            message = invalid_argument_template.format(
                variables=", ".join(map(repr, sorted(errors))),
                cls=self.cls_name,
                method=self.name,
                violations=format_violations(kwargs, errors),
                contract=self.format_contract_fields(errors),
            )
            raise ContextContractError(message)
        return result

    def check_success_statement(self, method, ctx, ns):
        __tracebackhide__ = True
        super(SpecContract, self).check_success_statement(method, ctx, ns)
        unknown = self.identify(ns)
        if unknown:
            message = unknown_variable_template.format(
                unknown=", ".join(map(repr, sorted(unknown))),
                cls=method.__self__.__class__.__name__,
                method=method.__name__,
                contract=self,
            )
            raise ContextContractError(message)
        kwargs, errors = self.validate(ns)
        if errors:
            message = invalid_variable_template.format(
                variables=", ".join(map(repr, sorted(errors))),
                cls=method.__self__.__class__.__name__,
                method=method.__name__,
                violations=format_violations(ns, errors),
                contract=self.format_contract_fields(errors),
            )
            raise ContextContractError(message)
        return kwargs

    def identify(self, ns):
        available = set(self.spec) | set(self.argset)
        unknown = set(ns) - available
        return unknown

    def validate(self, ns):
        __tracebackhide__ = True
        result, errors, seen, conflict = {}, {}, [], {}
        for key, value in ns.items():
            if key in self.spec:
                self.validate_spec(result, errors, seen, key, value)
            else:
                self.validate_argset(result, errors, seen, conflict, key, value)
        if conflict:
            conflict_vars = sorted({j for i in conflict.values() for j in i})
            message = normalization_conflict_template.format(
                conflict=", ".join(map(repr, conflict_vars)),
                results="\n\n".join(
                    "%s.%s:\n%s"
                    % (
                        cls,
                        method,
                        "\n".join(" - %s: %r" % (i, result[i]) for i in sorted(result)),
                    )
                    for (cls, method), result in (
                        (i, conflict[i]) for i in sorted(conflict)
                    )
                ),
                contract=self.format_contract_fields(conflict_vars),
            )
            raise ContextContractError(message)
        return result, errors

    def validate_spec(self, result, errors, seen, key, value):
        new_value, error = self.spec[key](value)
        if error:
            errors[key] = error
        else:
            self.assign_result(result, seen, key, value, new_value)

    def validate_argset(self, result, errors, seen, conflict, key, value):
        new_values, has_error = [], False
        for validator, cls_name, name in self.argset[key]:
            new_value, error = validator(value)
            if error:
                has_error = True
                errors[key] = error
            else:
                new_values.append((new_value, cls_name, name))
        if new_values:
            first, others = new_values[0], new_values[1:]
            for other in others:
                if first[0] != other[0]:
                    conflict.setdefault((first[1], first[2]), {})
                    conflict[(first[1], first[2])][key] = first[0]
                    conflict.setdefault((other[1], other[2]), {})
                    conflict[(other[1], other[2])][key] = other[0]
        if not has_error:
            self.assign_result(result, seen, key, value, new_value)

    def assign_result(self, result, seen, key, value, new_value):
        for seen_key, seen_value in seen:
            if value is seen_value:
                seen_new_value = result[seen_key]
                if (
                    type(new_value) is type(seen_new_value)
                    and new_value == seen_new_value
                ):
                    result[key] = seen_new_value
                    return
        result[key] = new_value
        seen.append((key, value))

    def __repr__(self):
        return self.format_contract_fields(self.argset, self.declared)

    def format_contract_fields(self, *fieldset):
        lines = ["Contract:"]
        arguments = sorted(
            [field for fields in fieldset for field in fields if field in self.argset]
        )
        for argument in arguments:
            validators = self.argset[argument]
            if len(validators) == 1:
                ((validator, cls_name, name),) = validators
                lines.append(
                    "  %s: %r  # Argument of %s.%s"
                    % (argument, validator, cls_name, name)
                )
            else:
                lines.append("  %s:" % (argument,))
                for validator in sorted(validators, key=itemgetter(1, 2)):
                    lines.append("    %r  # Argument of %s.%s" % validator)
        variables = sorted(
            [field for fields in fieldset for field in fields if field in self.declared]
        )
        for variable in variables:
            cls_name, name, field_name = self.declared[variable]
            lines.append(
                "  %s: %s  # Variable in %s.%s" % (variable, field_name, cls_name, name)
            )
        return "\n".join(lines)


def format_violations(ns, errors):
    result = []

    def normalize(value, indent, list_item=False, dict_value=False):
        if isinstance(value, dict):
            normalize_dict(value, indent + 2)
        elif isinstance(value, list):
            normalize_list(value, indent if list_item else indent + 2)
        elif isinstance(value, PydanticError):
            indent = indent + 2 if dict_value else indent
            normalize_pydantic(value, indent)
        else:
            indent = indent + 2 if dict_value else indent
            normalize_str(value, indent)

    def normalize_dict(value, indent, sep=None):
        for key in sorted(value):
            normalize(str(key) + ":", indent)
            if sep is not None:
                normalize(repr(ns[key]), indent, dict_value=True)
            normalize(value[key], indent, dict_value=True)
            if sep is not None:
                normalize_str(sep, 0)

    def normalize_list(value, indent):
        for elem in value:
            normalize(elem, indent, list_item=True)

    def normalize_pydantic(value, indent):
        normalize_str(str(value.exc), indent)

    def normalize_str(value, indent):
        result.append(" " * indent + value)

    normalize_dict(errors, 0, "")

    return "\n".join(result).strip()


# Wrap.


def combine_contract(parent, child):
    if type(parent) is NullContract and type(child) is NullContract:
        combine_argsets(parent, child)
        return
    elif (
        type(parent) is SpecContract
        and type(child) is SpecContract
        and parent.origin is child.origin
    ):
        combine_argsets(parent, child)
        return
    elif (
        type(parent) is SpecContract
        and type(child) is SpecContract
        and any(
            isinstance(parent.origin, spec_type) and isinstance(child.origin, spec_type)
            for spec_type in [PydanticSpec, MarshmallowSpec, CerberusSpec, dict]
        )
    ):
        repeated = set(parent.declared) & set(child.declared)
        if repeated:
            # FIXME: Repeated variables can occur in three different
            # classes.
            key = next(iter(repeated))
            message = incompatible_contracts_template.format(
                repeated=", ".join(map(repr, sorted(repeated))),
                cls=parent.declared[key][0],
                method=parent.declared[key][1],
                other_cls=child.declared[key][0],
                other_method=child.declared[key][1],
            )
            raise ContextContractError(message)
        combine_argsets(parent, child)
        combine_declared(parent, child)
    else:
        message = type_error_template.format(
            cls=parent.cls_name,
            method=parent.name,
            contract=format_contract(parent),
            other_cls=child.cls_name,
            other_method=child.name,
            other_contract=format_contract(child),
        )
        raise ContextContractError(message)


def combine_argsets(parent, child):
    for key in set(parent.argset) & set(child.argset):
        parent.argset[key].update(child.argset[key])
        child.argset[key].update(parent.argset[key])
    for key in set(parent.argset) - set(child.argset):
        child.argset[key] = parent.argset[key]
    for key in set(child.argset) - set(parent.argset):
        parent.argset[key] = child.argset[key]


def combine_declared(parent, child):
    parent.declared.update(child.declared)


def format_contract(contract):
    if type(contract) is SpecContract:
        if isclass(contract.origin):
            return contract.origin.__bases__[0]
        else:
            return type(contract.origin)
    else:
        return None


def maybe_extend_downstream_argsets(methods, root):
    if type(root) is NullContract:
        return
    for _method, contract, _protocol in methods:
        combine_argsets(root, contract)


# Messages.


undefined_argument_template = """
These arguments should be declared in the context contract: {undefined}

Story method: {cls}.{method}

Story arguments: {arguments}
""".strip()


missed_variable_template = """
These variables are missing from the context: {missed}

Story method: {cls}.{method}

Story arguments: {arguments}

{ctx!r}
""".strip()


variable_override_template = """
These variables are already present in the context: {variables}

Function returned value: {cls}.{method}

Use different names for Success() keyword arguments.

{ctx!r}
""".strip()


unknown_variable_template = """
These variables were not defined in the context contract: {unknown}

Function returned value: {cls}.{method}

Use different names for Success() keyword arguments or add these names to the contract.

{contract!r}
""".strip()


unknown_argument_template = """
These arguments are unknown: {unknown}

Story method: {cls}.{method}

{contract!r}
""".strip()


invalid_variable_template = """
These variables violates context contract: {variables}

Function returned value: {cls}.{method}

Violations:

{violations}

{contract}
""".strip()


invalid_argument_template = """
These arguments violates context contract: {variables}

Story method: {cls}.{method}

Violations:

{violations}

{contract}
""".strip()


incompatible_contracts_template = """
Repeated variables can not be used in a story composition.

Variables repeated in both context contracts: {repeated}

Story method: {cls}.{method}

Substory method: {other_cls}.{other_method}

Use variables with different names.
""".strip()


type_error_template = """
Story and substory context contracts has incompatible types:

Story method: {cls}.{method}

Story context contract: {contract}

Substory method: {other_cls}.{other_method}

Substory context contract: {other_contract}
""".strip()


normalization_conflict_template = """
These arguments have normalization conflict: {conflict}

{results}

{contract}
""".strip()



from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import NoReturn
from typing import Union

from _stories.contract import NullContract
from _stories.contract import SpecContract
from _stories.history import History

def make_context(
    contract: Union[SpecContract, NullContract],
    kwargs: Dict[str, Any],
    history: History,
) -> Context: ...

class Context:
    def __getattr__(self, name: str) -> Any: ...
    def __setattr__(self, name: str, value: int) -> NoReturn: ...
    def __delattr__(self, name: str) -> NoReturn: ...
    def __repr__(self) -> str: ...
    def __dir__(self) -> List[str]: ...
    def __bool__(self) -> NoReturn: ...

def assign_namespace(
    ctx: Context, method: Callable, kwargs: Dict[str, Any]
) -> None: ...
def history_representation(ctx: Context) -> str: ...
def context_representation(ctx: Context, repr_func: Callable = ...) -> str: ...



from _stories.mounted import ClassMountedStory


def contract_in(cls, *args):
    def setter(contract):
        for attrname in dir(cls):
            attribute = getattr(cls, attrname)
            if type(attribute) is ClassMountedStory:
                attribute.contract(contract)
        return contract

    if args:
        return setter(*args)
    else:
        return setter


def failures_in(cls, *args):
    def setter(failures):
        for attrname in dir(cls):
            attribute = getattr(cls, attrname)
            if type(attribute) is ClassMountedStory:
                attribute.failures(failures)
        return failures

    if args:
        return setter(*args)
    else:
        return setter
